#!/usr/local/bin/perl
#
# Orbitus Mesh Model Processor
# Author: Douglas Hawkinson
# Date: 05-21-06
#
# Usage: preproc.pl
#
# OutFile: *.orbitusT This is a table file to build containters off of
# 	     *.orbitusM This is a mesh file to read in actual objects
#

my $out_table = ".orbitusT";
my $out_mesh  = ".orbitusM";
my $out_data = "meshes.data";
my $self = "preproc.pl";
my $path = "..\\data\\meshes\\";
my $dir = '.';
my $uid = 0;

# Init the needed vars


my @files;

opendir ( DH, $dir ) or die "Failed to open $dir: $!";

#@files = grep (!/^((\.)+)|(.svn)|($self)|($out_table)|($out_mesh)$/, readdir (DH));

@files = grep ( /\.[xX]/, readdir( DH ));

# Build table files
# Will only build a table file for .X files with more than one
#  object.  The parser will run through the file and pull out
#  all the object names and output them to a .orbitusT file

my @objects = 0;
my @trans = 0;
my $counter = 0;
my $counter2 = 0;
my $t_line = 0;

my @AoA = 0;
my @AoA2 = 0;

print "* Opening meshes.data for writing...\n";
open DATA_O, "meshes.data";

print "\n=== Building Orbitus table files ===\n";

foreach( @files ) {

    # open file
    print "* Opening $_ for reading... \n";
    open X_FILE, "$_";
    @input = <X_FILE>;

    $counter = 0; # counter for number of objects

    $stage_t = 0; # stage for tables

    # Run throug file
    TABLE:foreach( @input ) {

	#if( /FrameTransformMatrix/ ) {
	#    $stage = 1;
	#}

	# Find Frame or Mesh with a name appened to it
	if( /(Frame|Mesh)\s[a-zA-Z0-9\-_]+\s/ ) {
	    
	    #print "$_ :::";

	    @obj = split /\s/;

	    #print "OBJECT: @obj[0] @obj[1] @obj[2]\n";

	    if( @obj[2] =~ /[0-9]+/ ) {
		print "IN: ";
		@obj = split /[0-9]+/, @obj[2];
	    }
	    else {
		@obj[0] = @obj[2];
	    }
	    print "@obj\n";
	    
	    @objects[$counter] = @obj[0];
	    
	    #print " @obj[0]\n";
	    $counter++;
	    #$counter2++;
	    
	    $stage = 1;

	    $temp_obj = @obj[0];

	    @obj = 0;

       	}
	elsif( $stage == 1 && /FrameTransformMatrix/ ) {

	    if( $temp_obj eq "TestCube" ) {
		print "Stage1: TestCube $_\n";
		print "------ @input[$t_line + 1]";
	    }
	    
	    @trans_num = split /,/, @input[$t_line + 1];

	    # Get rid of the garbage spacing
	    @temp = split /\s/, @trans_num[0];
	    
	    foreach( @temp ) {
		if( $_ =~ /[0-9]+/ ) {
		    @trans_num[0] = $_;
		}
	    }

	    @trans_num[15] = split /;;/, @trans_num[15]; # Remove ;;
	    
	    for $x ( 0 ... 15 ) {
		$AoA[$counter-1][$x] = @trans_num[$x];
	    }

	    $stage = 2;

	}
	elsif( $stage == 2 && /FrameTransformMatrix/ ) {

	    if( $temp_obj eq "TestCube" ) {
		print "Stage2: TestCube $_\n";
		print "------ @input[$t_line + 1]";
	    }

	    @trans_num = split /,/, @input[$t_line + 1];

	    # Get rid of the garbage spacing
	    @temp = split /\s/, @trans_num[0];

	    foreach( @temp ) {
		if( $_ =~ /[0-9]+/ ) {
		    @trans_num[0] = $_;
		}
	    }

	    @trans_num[15] = split /;;/, @trans_num[15]; # Remove ;;
	    
	    for $x ( 0 ... 15 ) {
		$AoA2[$counter-1][$x] = @trans_num[$x];
	    }

	    $stage = 0;

	}

	$t_line++;
    }

    # Go through objects array if greater than 1 and build the table file
    if( $counter > 1 ) {
	
	# Construct output file name and open it
	@out_target = split /\./;
	$out_file = @out_target[0] . $out_table;
	print "** Constructing $out_file...\n";
	open OUT, "> $out_file";
	
	# Print out containter
	@temp = @trans[0];
	print OUT "P @objects[0]\n";

	for( $i = 1; $i < $counter; $i++ ) {

	    @t_temp = @trans[$i];
	    #print "C @objects[$i] @{$AoA[$i]} @{$AoA2[$i]}\n";
	    print OUT "C @objects[$i] @{$AoA[$i]} @{$AoA2[$i]}\n";
	
	}

	close OUT;

    }
    $t_line = 0;
    @AoA = 0;
    @AoA2 = 0;
    # Close .X file
    close X_FILE;

}
    
# Build mesh files
# Do multiple passes on a file to build our own mesh format for easier and
#  faster reading in the actual program.

print "\n=== Building Orbitus Mesh files ===\n";

# Texuture Name Caching
my $num_tex_name = 0;
my @tex_names = 0;
# Vertex Caching
my $num_vert = 0;
my @vertex_arr_x = 0;
my @vertex_arr_y = 0;
my @vertex_arr_z = 0;
# Index Caching
my $num_ind = 0;
my @index_arr_x = 0;
my @index_arr_y = 0;
my @index_arr_z = 0;
# Transform Cache
my @t_num = 0;
# Mesh Normal Caching
my $num_normal = 0;
my @normal_x = 0;
my @normal_y = 0;
my @normal_z = 0;
# Mesh Normal Index Caching
my $num_nindex = 0;
my @nindex_x = 0;
my @nindex_y = 0;
my @nindex_z = 0;
# Texture map list
my @num_tex = 0;
my @textures = 0;
# Texture Coord Caching
my @num_texc = 0;
my @texCoord_u = 0;
my @texCoord_v = 0;
# Cached names of created objects
my @created_objs;
my $num_created = 0;
my $name = 0;

my $cached_line = 0;

# open data file
my $uid = 0;
print " Opening meshes.data for writing... \n";
open OUT_D, "> $out_data";

foreach( @files ) {

    # open input file
    print "\n* Opening $_ for reading... \n";
    open X_FILE, "$_";
    @input = <X_FILE>;

    $line = 0;  # line number
    $cached_line = 0;
    $stage = 0; # stage number, for determining which ifs to ignore

    # Run through file
    INPUT:foreach( @input ) {

	$cached_line = $line;

	# Stage 0.a; I know 0.a is bad but I dont feel like redoing the numbering it
	if( /TextureFilename\s/ && $stage < 10 ) {
	    
	    @t_name = split /\"/, @input[$line + 1];
	    @t_name = @t_name[1];

	    @tex_names[$num_tex_name] = @t_name;
	    $num_tex_name++;

	}
	    

	# Stage 0: Find Frame or Mesh with a name appended to it
	if( $stage == 0 &&
	    (/Frame\s[a-zA-Z0-9\-_]+\s/ || /Mesh\s[a-zA-Z0-9\-_]+\s/) ) {
	    
	    @obj = split /\s/;
	    $obj_name = @obj[1];
	    $stage++;

	    $s_name = @obj[1];

	    if( $obj_name eq "Frame " || $obj_name eq "Frame" || $obj_name eq " Frame" ) {
		print "@obj[2]\n";
		$obj_name = @obj[2];
		$name = @obj[2];
		$s_name = @obj[2];
	    }

	    # check against objects already created to see if anything more than
	    # the transformation matrix needs to be stored
	    if( $obj_name =~ /[0-9]+/ ) {
		
		@obj_name = split /[0-9]+/, $obj_name;

		# Find the text
		foreach( @obj_name ) {
		    if( /^[a-zA-Z_]+$/ ) {
			$name = $_;
		    }
		}
		#$name = @obj_name[0];
		
		foreach( @created_objs ) {
		    if( $_ eq $name ) {
			print "!!! Throwing out $obj_name $name\n";
			$stage = 20;
			$line = $cachedline + 1;
			next INPUT;
		    }
		}

		print "** Creating new object ## $name ...\n";
		push (@created_objs, $name);
		$obj_name = $name;
		#@created_objs[$num_created] = $name;
		$num_created++;		

	    }
	    elsif( $obj_name =~ /^[a-zA-Z_]+$/ ) {

		print "** Creating new object XX $obj_name $s_name ...\n";
		$name = $obj_name;
		push (@created_objs, $s_name);
		#@created_objs[$num_created] = $obj_name;
		$num_created++;	

	    }
	    else {
		print "Illegal object, skipping...\n";
		$stage = 20;
	    }
	}
	# Stage 1: Find the FrameTransformation Matrix
	elsif( $stage == 1 && /FrameTransformMatrix/ ) {
	 
	    @t_num = split /,/, @input[$line + 1];

	    # Get rid of the garbage spacing
	    @temp = split /\s/, @t_num[0];
	    
	    foreach( @temp ) {
		if( $_ =~ /[0-9]+/ ) {
		    @t_num[0] = $_;
		}
	    }

	    @t_num[15] = split /;;/, @t_num[15]; # Remove ;;
	    
	    $stage++;

	}
	# Stage 2: Find Mesh, and read in vertices
	elsif( $stage == 2 && /Mesh\s/ ) {
	    
	    @num = split /;/, @input[$line + 1];
	    @num = split /\s/, @num[0];

	    # Get rid of the garbage spacing
	    foreach( @num ) {
		if( $_ =~ /[0-9]+/ ) {
		    $num_vert = $_;
		}
	    }

	    $line++;

	    $temp_ctr = $line + 1;

	    # Stage2.2: read in Vetices
	    for( $i = $line + 1; $i <= $line + $num_vert; $i++ ) {

		$line++;

	   	# Split vertices up
		@verts = split /;/, @input[$i];	
	
		# Get rid of the garbage spacing
		@temp = split /\s/, @verts[0];

		foreach( @temp ) {
		    if( $_ =~ /[0-9]+/ ) {
			@verts[0] = $_;
		    }
		}

		# Cache vertices for later writing
		@vertex_arr_x[$i - $temp_ctr] = @verts[0];
		@vertex_arr_y[$i - $temp_ctr] = @verts[1];
		@vertex_arr_z[$i - $temp_ctr] = @verts[2];
		
		# Catches final case
		if( @input[$i] =~ /;;/ ) {

		    $stage = 3;
		    #$line = $cached_line + 1;
		    $line = $cached_line + 1;
		    next INPUT;

		}
	    }
	}
	# Stage 3: Read in Indicies
	elsif( $stage == 3 && /;;/ ) {

	    @num = split /;/, @input[$line + 1];
	    @num = split /\s/, @num[0];

	    # Get rid of the garbage spacing
	    foreach( @num ) {
		if( $_ =~ /[0-9]+/ ) {
		    $num_ind = $_;
		}
	    }

	    $line++;

	    $temp_ctr = $line + 1;

	    # Stage 3.2: read in indices
	    for( $i = $line + 1; $i <= $line + $num_ind; $i++ ) {

		# Split the indices up
		@inds = split /;/, @input[$i];
		@inds = split /,/, @inds[1];

		# Cache indices for later writing
		@index_arr_x[$i - $temp_ctr] = @inds[0];
		@index_arr_y[$i - $temp_ctr] = @inds[1];
		@index_arr_z[$i - $temp_ctr] = @inds[2];
		
		# catches final case
		if( @input[$i] =~ /;;/ ) {
		    
		    $stage = 4;
		    $line = $cached_line + 1;
		    next INPUT;

		}
	    }
	}
	# Stage 4: Read in Normals
	elsif( $stage == 4 && /MeshNormals\s/ ) {

	    @num = split /;/, @input[$line + 1];
	    @num = split /\s/, @num[0];

	    # Get rid of the garbage spacing
	    foreach( @num ) {
		if( /[0-9]+/ ) {
		    $num_normal = $_;
		}
	    }

	    $line++;
	    $temp_ctr = $line + 1;
	    
	    for( $i = $line + 1; $i <= $line + $num_normal; $i++ ) {
		
		$line++;

	   	# Split vertices up
		@norms = split /;/, @input[$i];	
		
		# Get rid of the garbage spacing
		@temp = split /\s/, @norms[0];

		foreach( @temp ) {
		    if( /[0-9]+/ ) {
			$norms[0] = $_;
		    }
		}

		# Cache vertices for later writing
		@normal_x[$i - $temp_ctr] = @norms[0];
		@normal_y[$i - $temp_ctr] = @norms[1];
		@normal_z[$i - $temp_ctr] = @norms[2];

		if( @input[$i] =~ /;;/ ) {
		    $stage = 5;
		    $line = $cached_line + 1;
		    next INPUT;
		}
	    }
	}
	# Stage 5: Read in Normal Indices
	elsif( $stage == 5 && /;;/ ) {

	    @num = split /;/, @input[$line + 1];
	    @num = split /\s/, @num[0];

	    # Get rid of the garbage spacing
	    foreach( @num ) {
		if( $_ =~ /[0-9]+/ ) {
		    $num_nindex = $_;
		}
	    }

	    $line++;

	    $temp_ctr = $line + 1;

	    # Stage 5.2: read in indices
	    for( $i = $line + 1; $i <= $line + $num_nindex; $i++ ) {

		# Split the indices up
		@ninds = split /;/, @input[$i];
		@ninds = split /,/, @ninds[1];

		# Cache indices for later writing
		@nindex_x[$i - $temp_ctr] = @ninds[0];
		@nindex_y[$i - $temp_ctr] = @ninds[1];
		@nindex_z[$i - $temp_ctr] = @ninds[2];
		
		# catches final case
		if( @input[$i] =~ /;;/ ) {
		    
		    $stage = 6;
		    $line = $cached_line + 1;
		    next INPUT;

		}
	    }	    
	}
	# Stage 6: Read in Mesh Material List
	elsif( $stage == 6 && /MeshMaterialList\s/ ) {

	    @num = split /;/, @input[$line + 2];
	    @num = split /\s/, @num[0];

	    # Get rid of garbage spacing
	    foreach( @num ) {
		if( /[0-9]+/ ) {
		    $num_tex = $_;
		}
	    }

	    $line++;

	    $temp_ctr = $line + 2;

	    # Stage 6.2: Read in map numbers
	    for( $i = $line + 2; $i <= $line + $num_tex + 1; $i++ ) {

		# Get workable number
		if( @input[$i] =~ /,/ ) {
		    @tex = split /,/, @input[$i];
		}
		elsif( @input[$i] =~ /;/ ) {
		    @tex = split /;/, @input[$i];
		}

		@tex = split /\s/, @tex[0];

		# Cache mesh material
		foreach( @tex ) {
		    if( /[0-9]+/ ) {
			@tex = $_;
		    }
		}
		@textures[$i - $temp_ctr] = @tex;

		# catches final case
		if( @input[$i] =~ /;/ ) {

		    $stage = 7;
		    $line = $cached_line + 1;
		    next INPUT;

		}
	    }
	}
	# Stage 7: Read in Texture Coordinate data
	elsif( $stage == 7 && /MeshTextureCoords\s/ ) {

	    @num = split /;/, @input[$line + 1];
	    @num = split /\s/, @num[0];

	    # Get rid of garbage spacing
	    foreach( @num ) {
		if( /[0-9]+/ ) {
		    $num_texc = $_;
		}
	    }

	    $line++;
	    
	    $temp_ctr = $line + 1;

	    # Stage 7.2: Read in numbers
	    for( $i = $line + 1; $i <= $line + $num_texc; $i++ ) {

		# Split vertices up
		@texc = split /;/, @input[$i];	
		
		# Get rid of the garbage spacing
		@temp = split /\s/, @texc[0];
		
		foreach( @temp ) {
		    if( /[0-9]+/ ) {
			$texc[0] = $_;
		    }
		}

		# Cache vertices for later writing
		@texCoord_u[$i - $temp_ctr] = @texc[0];
		@texCoord_v[$i - $temp_ctr] = @texc[1];

		if( @input[$i] =~ /;;/ ) {
		    $stage = 8;
		    $line = $cached_line + 1;
		    next INPUT;
		}

	    }

	}
	# Stage 8: Write out data
	elsif( $stage == 8 ) {

	    # open mesh file
	    $out_file_m = $obj_name . $out_mesh;
	    print "*** Constructing $out_file_m...\n";
	    open OUT_M, "> $out_file_m";

	    # Write out object name
	    print OUT_M "$obj_name\n";
	    print OUT_D "$uid $obj_name $path $obj_name" . ".orbitusM\n";
	    $uid++;

	    # Write transform matrix in four rows
	    for( $a = 0; $a < 16; $a += 4 ) {
		print OUT_M "@t_num[$a] @t_num[$a+1] @t_num[$a+2] @t_num[a+3]\n";
	    }
	    
	    # Write texture names
	    print OUT_M "Textures\n";
	    print OUT_M "$num_tex_name\n";

	    $ctr = 0;

	    foreach( @tex_names ) {
		print OUT_M "$ctr $_\n";
		$ctr++;
	    }

	    # Write the Vertices and texture coords
	    print OUT_M "Vertices\n";
	    print OUT_M "$num_vert\n";
	    
	    $ctr = 0;

	    foreach( @vertex_arr_x ) {
		# vertices
		print OUT_M "$_ @vertex_arr_y[$ctr] @vertex_arr_z[$ctr] ";
		# texture coordinates
		print OUT_M "@texCoord_u[$ctr] @texCoord_v[$ctr]\n";
	        $ctr++;
	    }

	    # Write out the Indices
	    print OUT_M "Indices\n";
	    print OUT_M "$num_ind\n";

	    $ctr = 0;

	    foreach( @index_arr_x ) {
		print OUT_M "$_ @index_arr_y[$ctr] @index_arr_z[$ctr]\n";
		$ctr++;
	    }

	    # Write out the Normals
	    print OUT_M "MeshNormals\n";
	    print OUT_M "$num_normal\n";

	    $ctr = 0;

	    foreach( @normal_x ) {
		print OUT_M "$_ @normal_y[$ctr] @normal_z[$ctr]\n";
		$ctr++;
	    }

	    # Write out the Normal Indices
	    print OUT_M "NormalIndices\n";
	    print OUT_M "$num_nindex\n";

	    $ctr = 0;

	    foreach( @nindex_x ) {
		print OUT_M "$_ @nindex_y[$ctr] @nindex_z[$ctr]\n";
		$ctr++;
	    }

	    # Write out the texture map list
	    print OUT_M "TextureMap\n";
	    print OUT_M "$num_tex\n";
	    
	    $ctr = 0;
	    
	    foreach( @textures ) {
		print OUT_M "$_\n";
	    }

	    # Close mesh file
	    close OUT_M;
	    
	    $stage = 20;

	}
	
	if( ( $stage != 0 && /(Frame|Mesh)\s[a-zA-Z0-9\-_]+{/ )  || $stage == 20 )  {
	    	    
            # Clean up crap
	    @vertex_arr_x = 0;
	    @vertex_arr_y = 0;
	    @vertex_arr_z = 0;
	    @index_arr_x = 0;
	    @index_arr_y = 0;
	    @index_arr_z = 0;
	    @normal_x = 0;
	    @normal_y = 0;
	    @normal_z = 0;
	    @nindex_x = 0;
	    @nindex_y = 0;
	    @nindex_z = 0;
	    @textures = 0;
	    @texCoord_u = 0;
	    @texCoord_v = 0;
	    @tex_names = 0;
	    $num_tex_name = 0;
	    $num_tex = 0;
	    $num_nindex = 0;
	    $num_normal = 0;
	    $num_ind = 0;
	    $num_vert = 0;
	    @t_num = 0;

	    $stage = 0;

	}

	$line++;

    }

}
# close files
close OUT_M;
close OUT;
close OUT_D;	
close X_FILE;

closedir DH;
 
